/*
 * Copyright (C) 2008 Intel Corporation. All Rights Reserved.
 *
 * The source code contained or described herein and all
 * documents related to the source code ("Material") are owned by
 * Intel Corporation or its suppliers or licensors. Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors. The Material is protected by worldwide copyright
 * laws and treaty provisions.  No part of the Material may be
 * used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed,  or disclosed in any way
 * except as expressly provided in the license provided with the
 * Materials.  No license under any patent, copyright, trade
 * secret or other intellectual property right is granted to or
 * conferred upon you by disclosure or delivery of the Materials,
 * either expressly, by implication, inducement, estoppel or
 * otherwise, except as expressly provided in the license
 * provided with the Materials
*/
// [DESCRIPTION]
// 
// Solve the nqueens problem openmp version
//
// [COMPILE]  
// Compile on IA-32 and Intel(R) 64 architectures only. 
// Use the following compiler options to compile: 
//
// CommandLine: icl /Qopenmp nq-openmp.cpp winmm.lib
//  

#include <iostream>  // std::cout, cin, cerr ...
#include <iomanip>   // modify std::out
#include <time.h>


#include"omp.h"
#define _NOSYNC
//#define OMP
//#define DEBUG
using namespace std;
#ifdef _NOSYNC
#define OMP
#endif
#ifdef _NOSYNC
#define MAX_THREAD 2 // Modify if you wish to use more
int solcnt[MAX_THREAD];
#endif
int nrOfSolutions=0;
int size=0;

#include <windows.h>
#include <mmsystem.h>

void print(int queens[]) {
  cerr << "Solution " << nrOfSolutions << endl; 
  for(int row=0; row<size; row++) {
    for(int col=0; col<size; col++) {
      if(queens[row]==col) {
        cout << "Q";
      }
      else {
        cout << "-";
      }
    }
    cout << endl;
  }
}

void setQueen(int queens[], int row, int col, int id) {
  
  for(int i=0; i<row; i++) {
    // vertical attacks
    if (queens[i]==col) {
      return;
    }
    // diagonal attacks
    if (abs(queens[i]-col) == (row-i) ) {
      return;
    }
  }

  // column is ok, set the queen
  queens[row]=col;

  if(row==size-1) {
    // only one thread should print allowed to print at a time
    // increasing the solution counter is not atomic
#ifdef _NOSYNC
    solcnt[id]++;
#else
#pragma omp atomic 
    nrOfSolutions++;
#endif 
#ifdef DEBUG
#pragma omp critical
    print(queens);
#endif
  }
  else {
    // try to fill next row
    for(int i=0; i<size; i++) {
      setQueen(queens, row+1, i, id);
    }
  }
}


void solve() {
  int myid ;
#pragma omp parallel for private(myid)
  for(int i=0; i<size; i++) {
#ifdef OMP
    myid = omp_get_thread_num(); 
    cout << "Thread number : " << myid << endl; 
#endif
    // try all positions in first row
    // create separate array for each recursion
    // started here
    setQueen(new int[size], 0, i, myid);
    }
#ifdef _NOSYNC
  for(int i=0; i<2; i++) { 
  nrOfSolutions += solcnt[i] ; 
  }
#endif
}


int main(int argc, char*argv[]) {
  
  if(argc !=2) {
    cerr << "Usage: nq-openmp boardSize.\n";
    return 0;
  }
  
  size = atoi(argv[1]);
  cout << "Starting OpenMP solver for size " << size << "...\n";

  DWORD startTime=clock();
  solve();
  DWORD endTime=clock();
  
  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took " << endTime-startTime << "ms.\n";
  
  return 0;
}
